package ru.dbotthepony.mc.otm.android

import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraftforge.registries.ForgeRegistryEntry
import ru.dbotthepony.mc.otm.capability.android.AndroidCapability

open class AndroidFeatureType<T : AndroidFeature> : ForgeRegistryEntry<AndroidFeatureType<*>> {
	private val factory: (AndroidFeatureType<T>, AndroidCapability) -> T

	constructor(factory: (AndroidCapability) -> T) : super() {
		this.factory = { _, capability -> factory.invoke(capability) }
	}

	constructor(factory: (AndroidFeatureType<T>, AndroidCapability) -> T) : super() {
		this.factory = factory
	}

	fun create(android: AndroidCapability): T {
		return factory.invoke(this, android)
	}

	open fun isApplicable(android: AndroidCapability) = true

	open val displayName: Component by lazy {
		TranslatableComponent(displayId)
	}

	open val displayId: String by lazy {
		registryName?.let { "android_feature.${it.namespace}.${it.path}" } ?: "android_feature.null.null"
	}

	companion object {
		var IS_REGISTRY_INITIALIZED = false
			private set

		fun markRegistryLoaded() {
			IS_REGISTRY_INITIALIZED = true
		}

		fun assert() {
			check(IS_REGISTRY_INITIALIZED) { "Registry is not yet initialized" }
		}

		inline fun assert(lazy: () -> Any) {
			check(IS_REGISTRY_INITIALIZED, lazy)
		}
	}
}