package ru.dbotthepony.mc.otm.capability.android

import net.minecraft.ChatFormatting
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraftforge.event.AttachCapabilitiesEvent
import net.minecraftforge.event.entity.player.PlayerEvent.Clone
import net.minecraftforge.event.entity.player.PlayerEvent.PlayerChangedDimensionEvent
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.ifPresentK
import ru.dbotthepony.mc.otm.network.*
import ru.dbotthepony.mc.otm.registry.AndroidFeatures
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.registry.MRegistry
import ru.dbotthepony.mc.otm.set
import java.util.*

class AndroidCapabilityPlayer(val ply: Player) : AndroidCapability(ply) {
	override var isAndroid = false
	private var remoteIsAndroid = false

	var willBecomeAndroid = false
	private var willBecomeAndroidNetwork = false
	private var shouldPlaySound = false
	private val research = ArrayList<AndroidResearch>()

	private var remoteBattery = ImpreciseFraction(-1)
	private var remoteMaxBattery = ImpreciseFraction(-1)
	private var remoteBatteryStack = ItemStack.EMPTY

	private var invalidateNetworkIn = 10

	fun invalidateNetworkState() {
		invalidateNetworkIn = 10
	}

	fun isEverAndroid(): Boolean = isAndroid || willBecomeAndroid

	fun becomeAndroidSoft() {
		if (isAndroid || willBecomeAndroid) return
		willBecomeAndroid = true
		(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message").withStyle(ChatFormatting.GRAY), false)
	}

	fun becomeAndroid() {
		if (isAndroid) return

		isAndroid = true
		willBecomeAndroid = false
		shouldPlaySound = false
		battery = ImpreciseFraction(60000)
		maxBattery = ImpreciseFraction(60000)
	}

	fun becomeAndroidAndKill() {
		if (isAndroid) return

		becomeAndroid()
		ply.hurt(MRegistry.DAMAGE_BECOME_ANDROID, ply.maxHealth * 2)
	}

	fun becomeHumane() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		isAndroid = false
		shouldPlaySound = false
		battery = ImpreciseFraction(0)
		maxBattery = ImpreciseFraction(60000)
		dropBattery()
	}

	fun becomeHumaneAndKill() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		becomeHumane()
		ply.hurt(MRegistry.DAMAGE_BECOME_HUMANE, ply.maxHealth * 2)
	}

	override fun deserializeNBT(compound: CompoundTag?) {
		super.deserializeNBT(compound!!)

		isAndroid = compound.getBoolean("is_android")
		willBecomeAndroid = compound.getBoolean("will_become_android")
		research.clear()

		val list = compound.getList("research", Tag.TAG_COMPOUND.toInt())

		for (tag in list) {
			if (tag is CompoundTag) {
				val research = MRegistry.ANDROID_RESEARCH.getValue(ResourceLocation(tag.getString("id")))

				if (research != null) {
					val instance = research.factory(this)
					instance.deserializeNBT(tag)
					this.research.add(instance)
				}
			}
		}
	}

	fun obliviate(refund: Boolean = true) {
		if (refund) {
			for (instance in research) {
				if (instance.isResearched) {
					instance.unResearch()
					instance.refund(simulate = false)
				}
			}
		}

		val copy = java.util.List.copyOf(features.values)

		for (feature in copy) {
			removeFeature(feature.type)
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag = super.serializeNBT()

		tag["is_android"] = isAndroid
		tag["will_become_android"] = willBecomeAndroid

		val list = ListTag()

		for (instance in research) {
			val researchTag = instance.serializeNBT()

			researchTag["id"] = instance.type.registryName!!.toString()
			list.add(researchTag)
		}

		tag["research"] = list
		return tag
	}

	fun <T : AndroidResearch> getResearch(type: AndroidResearchType<T>): T {
		for (instance in research) {
			if (instance.type === type) {
				return instance as T
			}
		}

		val instance = type.factory(this)
		research.add(instance)
		return instance
	}

	var lastJumpTicks = 14

	override fun tick() {
		super.tick()

		if (invalidateNetworkIn > 0 && --invalidateNetworkIn == 0) {
			remoteBattery = ImpreciseFraction.MINUS_ONE
			remoteMaxBattery = ImpreciseFraction.MINUS_ONE
			remoteBatteryStack = ItemStack.EMPTY

			remoteIsAndroid = false
			willBecomeAndroidNetwork = false

			for (instance in research) {
				instance.isDirty = true
				instance.invalidateNetwork()
			}

			for (feature in features.values) {
				feature.isDirty = true
				feature.invalidateNetwork()
			}
		}

		tickedOnce = true

		if (battery != remoteBattery) {
			remoteBattery = battery
			sendNetwork(AndroidEnergyLevelPacket(level = battery, isMaxEnergy = false))
		}

		if (maxBattery != remoteMaxBattery) {
			remoteMaxBattery = maxBattery
			sendNetwork(AndroidEnergyLevelPacket(level = maxBattery, isMaxEnergy = true))
		}

		if (!remoteBatteryStack.equals(batteryItemStack, false)) {
			remoteBatteryStack = batteryItemStack.copy()
			sendNetwork(AndroidBatteryItemPacket(batteryItemStack))
		}

		if (networkQueue.size != 0) {
			for (packet in networkQueue) {
				AndroidNetworkChannel.send(ply, packet)
			}

			networkQueue.clear()
		}

		if (isAndroid != remoteIsAndroid) {
			remoteIsAndroid = isAndroid
			sendNetwork(IsAndroidPacket(isAndroid))
			shouldPlaySound = false
		}

		if (willBecomeAndroid != willBecomeAndroidNetwork) {
			willBecomeAndroidNetwork = willBecomeAndroid
			sendNetwork(WillBecomeAndroidPacket(willBecomeAndroid))
		}

		for (instance in research) {
			if (instance.isDirty) {
				instance.isDirty = false
				sendNetwork(AndroidResearchSyncPacket(instance))
			}
		}

		for (instance in features.values) {
			if (instance.isDirty) {
				instance.isDirty = false
				sendNetwork(AndroidFeatureSyncPacket(instance))
			}
		}
	}

	var sleepTicks = 0

	override fun tickInnerClientAlways() {
		super.tickInnerClientAlways()

		if (willBecomeAndroid) {
			if (ply.isSleeping) {
				sleepTicks++
			} else {
				sleepTicks = 0
			}
		}
	}

	override fun tickServerAlways() {
		super.tickServerAlways()

		if (willBecomeAndroid) {
			if (ply.isSleeping) {
				sleepTicks++

				if (sleepTicks > SLEEP_TICKS_LIMIT) {
					becomeAndroid()
					shouldPlaySound = true
					sleepTicks = 0

					(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
				}
			} else {
				sleepTicks = 0
			}
		}
	}

	public override fun tickServer() {
		super.tickServer()

		// TODO: Maybe passive drain?
		// extractEnergyInner(BigDecimal.valueOf(new Random().nextDouble()), false);
		if (ply.isSwimming && !hasFeature(AndroidFeatures.AIR_BAGS)) {
			ply.isSwimming = false
		}

		val stats = ply.foodData

		while (stats.foodLevel < 18 && extractEnergyInner(ENERGY_FOR_HUNGER_POINT, true) >= ENERGY_FOR_HUNGER_POINT) {
			extractEnergyInner(ENERGY_FOR_HUNGER_POINT, false)
			stats.foodLevel = stats.foodLevel + 1
		}

		// "block" quick regeneration
		// also cause power to generate while in peaceful
		while (stats.foodLevel > 18 && receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, true) >= ENERGY_FOR_HUNGER_POINT) {
			receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, false)
			stats.foodLevel = stats.foodLevel - 1
		}

		val foodLevel = stats.foodLevel.toFloat()

		if (stats.saturationLevel < foodLevel) {
			val extracted = extractEnergyInner(ENERGY_FOR_HUNGER_POINT * (foodLevel - stats.saturationLevel), false)
			stats.setSaturation(stats.saturationLevel + (extracted / ENERGY_FOR_HUNGER_POINT).toFloat())
		}

		if (stats.exhaustionLevel > 0f) {
			val extracted = extractEnergyInner(ENERGY_FOR_HUNGER_POINT * (stats.exhaustionLevel / 4f), false)
			stats.setExhaustion(stats.exhaustionLevel - (extracted / ENERGY_FOR_HUNGER_POINT).toFloat() * 4f)
		}
	}

	@Suppress("unused")
	companion object {
		@SubscribeEvent
		fun onAttachCapabilityEvent(event: AttachCapabilitiesEvent<Entity?>) {
			val ent = event.`object`

			if (ent is Player) {
				event.addCapability(MNames.ANDROID_CAPABILITY_RS, AndroidCapabilityPlayer(ent))
			}
		}

		@SubscribeEvent
		fun onPlayerChangeDimensionEvent(event: PlayerChangedDimensionEvent) {
			event.player.getCapability(MatteryCapability.ANDROID)
				.ifPresentK { (it as AndroidCapabilityPlayer).invalidateNetworkState() }
		}

		@SubscribeEvent
		fun onPlayerCloneEvent(event: Clone) {
			val it = event.player.android as AndroidCapabilityPlayer? ?: return

			var resolver = event.original.getCapability(MatteryCapability.ANDROID)

			if (!resolver.isPresent || resolver.resolve().isEmpty) {
				event.original.reviveCaps()
				resolver = event.original.getCapability(MatteryCapability.ANDROID)
			}

			if (!resolver.isPresent || resolver.resolve().isEmpty) {
				event.original.invalidateCaps()
				return
			}

			val original = resolver.resolve().get() as AndroidCapabilityPlayer

			if (original.willBecomeAndroid && event.isWasDeath) {
				original.becomeAndroid()
				(event.player as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
			}

			it.deserializeNBT(original.serializeNBT())
			it.invalidateNetworkState()
			event.original.invalidateCaps()
		}

		val ENERGY_FOR_HUNGER_POINT = ImpreciseFraction(1000)
		const val SLEEP_TICKS_LIMIT = 80
	}
}