package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.resources.sounds.SimpleSoundInstance
import net.minecraft.network.chat.Component
import net.minecraft.sounds.SoundEvents
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.RenderHelper
import ru.dbotthepony.mc.otm.client.render.SkinElement
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.client.screen.panels.CheckBoxPanel.Companion.REGULAR_DIMENSIONS
import ru.dbotthepony.mc.otm.menu.widget.BooleanPlayerInputWidget

open class CheckBoxPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = REGULAR_DIMENSIONS,
	height: Float = REGULAR_DIMENSIONS
) : EditablePanel(
	screen, parent, x, y, width, height
) {
	open var checked = false

	override fun innerRender(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float) {
		if (checked) {
			CHECKBOX_CHECKED.render(stack)
		} else {
			CHECKBOX_UNCHECKED.render(stack)
		}
	}

	override fun mouseClickedInner(mouse_x: Double, mouse_y: Double, mouse_click_type: Int): Boolean {
		checked = !checked
		minecraft.soundManager.play(SimpleSoundInstance.forUI(SoundEvents.UI_BUTTON_CLICK, 1.0f))
		return true
	}

	companion object {
		const val REGULAR_DIMENSIONS = 15f

		val CHECKBOX_UNCHECKED = SkinElement(
			RenderHelper.WIDGETS,
			x = 18f,
			y = 65f,
			w = REGULAR_DIMENSIONS,
			h = REGULAR_DIMENSIONS,
		)

		val CHECKBOX_CHECKED = SkinElement(
			RenderHelper.WIDGETS,
			x = 18f,
			y = 80f,
			w = REGULAR_DIMENSIONS,
			h = REGULAR_DIMENSIONS,
		)
	}
}

open class CheckBoxInputPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	val widget: BooleanPlayerInputWidget,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = REGULAR_DIMENSIONS + 120f,
	height: Float = REGULAR_DIMENSIONS
) : CheckBoxPanel(screen, parent, x, y, width, height) {
	override var checked: Boolean
		get() = widget.value
		set(value) {}

	override fun mouseClickedInner(mouse_x: Double, mouse_y: Double, mouse_click_type: Int): Boolean {
		super.mouseClickedInner(mouse_x, mouse_y, mouse_click_type)
		widget.userInput(!checked)
		return true
	}
}

open class CheckBoxLabelPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	text: Component,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = REGULAR_DIMENSIONS + 120f,
	height: Float = REGULAR_DIMENSIONS
) : EditablePanel(screen, parent, x, y, width, height) {
	val checkbox = CheckBoxPanel(screen, this, 0f, 0f, REGULAR_DIMENSIONS, REGULAR_DIMENSIONS)
	val label = Label(screen, this, REGULAR_DIMENSIONS + 4f, 4f, text = text)
}

open class CheckBoxLabelInputPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	widget: BooleanPlayerInputWidget,
	text: Component,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = REGULAR_DIMENSIONS + 120f,
	height: Float = REGULAR_DIMENSIONS
) : EditablePanel(screen, parent, x, y, width, height) {
	val widget get() = checkbox.widget
	val checkbox = CheckBoxInputPanel(screen, this, widget, 0f, 0f, REGULAR_DIMENSIONS, REGULAR_DIMENSIONS)
	val label = Label(screen, this, REGULAR_DIMENSIONS + 4f, 4f, text = text)
}
