package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.ChatFormatting
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.client.render.RenderHelper
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.menu.MatterySlot
import javax.annotation.Nonnull

class SlotPanel<T : MatterySlot>(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	val slot: T,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 18f,
	height: Float = 18f
) : AbstractSlotPanel(screen, parent, x, y, width, height) {
	override fun mouseClickedInner(mouse_x: Double, mouse_y: Double, mouse_click_type: Int): Boolean {
		screen.onSlotClicked(slot, mouse_click_type)
		return true
	}

	override fun mouseDraggedInner(
		mouse_x: Double,
		mouse_y: Double,
		flag: Int,
		drag_x: Double,
		drag_y: Double
	): Boolean {
		screen.onSlotMouseDragged(slot, flag)
		return true
	}

	override fun mouseReleasedInner(mouse_x: Double, mouse_y: Double, flag: Int): Boolean {
		screen.onMouseReleasedSlot(slot, mouse_x, mouse_y, flag, false)
		return true
	}

	override fun getItemStack(): ItemStack {
		return slot.item
	}

	override fun innerRender(@Nonnull stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float) {
		renderSlotBackground(stack, mouse_x, mouse_y, flag)

		var itemstack = slot.item
		val carried = screen.menu.carried
		var dragHit = false
		var countOverride: String? = null

		if (!carried.isEmpty && screen.isQuickCrafting && screen.quickCraftSlots.contains(slot)) {
			if (screen.quickCraftSlots.size == 1) {
				return
			}

			if (AbstractContainerMenu.canItemQuickReplace(slot, carried, true) && screen.menu.canDragTo(slot)) {
				itemstack = carried.copy()
				dragHit = true

				AbstractContainerMenu.getQuickCraftSlotCount(
					screen.quickCraftSlots,
					screen.quickCraftingType,
					itemstack,
					if (slot.item.isEmpty) 0 else slot.item.count
				)

				val k = Math.min(itemstack.maxStackSize, slot.getMaxStackSize(itemstack))
				if (itemstack.count > k) {
					countOverride = ChatFormatting.YELLOW.toString() + k
					itemstack.count = k
				}
			} else {
				screen.quickCraftSlots.remove(slot)
				screen.recalculateQuickCraftRemaining()
			}
		}

		RenderSystem.setShader(GameRenderer::getPositionTexShader)

		if (dragHit) {
			RenderHelper.setDrawColor(SLOT_HIGHLIGHT_DRAG)
			RenderHelper.drawRect(stack, 1f, 1f, width - 1, height - 1)
		}

		renderRegular(stack, itemstack, countOverride)

		if (isHovered) {
			screen.hoveredSlot = slot
		}
	}

	override fun innerRenderTooltips(@Nonnull stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		// no op, screen does it for us (completely)
		return false
	}
}