package ru.dbotthepony.mc.otm.container

import net.minecraft.world.Container
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.capabilities.Capability
import ru.dbotthepony.mc.otm.set
import java.util.function.Consumer

class ContainerIteratorItemStack(private val container: Container) : MutableIterator<ItemStack>, Iterable<ItemStack> {
	private var scanIndex = 0
	private var nextStack: ItemStack? = null
	private var nextIndex = 0

	override fun hasNext(): Boolean {
		if (scanIndex >= container.containerSize && nextStack == null)
			return false

		while (nextStack == null && scanIndex < container.containerSize) {
			val get = container.getItem(scanIndex)

			if (!get.isEmpty) {
				nextStack = get
				nextIndex = scanIndex
			}

			scanIndex++
		}

		return nextStack != null
	}

	override fun next(): ItemStack {
		if (!hasNext()) {
			throw NoSuchElementException()
		}

		val nextStack = nextStack!!
		this.nextStack = null
		return nextStack
	}

	override fun remove() {
		container[nextIndex] = ItemStack.EMPTY
	}

	override fun iterator(): Iterator<ItemStack> {
		return this
	}

	fun consume(consumer: Consumer<ItemStack?>) {
		for (t in this) {
			consumer.accept(t)
		}
	}
}

fun Container.iterator() = ContainerIteratorItemStack(this)

class ContainerIteratorCapability<T>(private val container: Container, private val cap: Capability<T>) : Iterator<T>, Iterable<T> {
	private var index = 0
	private var nextStack: T? = null

	override fun hasNext(): Boolean {
		if (index >= container.containerSize && nextStack == null)
			return false

		while (nextStack == null && index < container.containerSize) {
			val get = container.getItem(index)

			if (!get.isEmpty) {
				get.getCapability(cap).ifPresent { t: T -> nextStack = t }
			}

			index++
		}

		return nextStack != null
	}

	override fun next(): T {
		val nextStack = nextStack!!
		this.nextStack = null
		return nextStack
	}

	override fun iterator(): Iterator<T> {
		return this
	}

	fun consume(consumer: Consumer<T>) {
		for (t in this) {
			consumer.accept(t)
		}
	}
}

fun <T> Container.iterator(cap: Capability<T>) = ContainerIteratorCapability(this, cap)
