package ru.dbotthepony.mc.otm.registry

import com.google.common.collect.ImmutableSet
import net.minecraft.world.item.DyeColor
import net.minecraftforge.registries.IForgeRegistryEntry
import net.minecraftforge.registries.RegistryObject

class LazyList<T>(private vararg val getters: () -> T) : AbstractList<T>() {
	override val size: Int
		get() = getters.size

	override fun get(index: Int): T {
		return getters[index].invoke()
	}
}

class LazyMap<K, T>(vararg mValues: Pair<K, () -> T>) : AbstractMap<K, T>() {
	private inner class Entry(
		override val key: K,
		private val getter: () -> T
	) : Map.Entry<K, T> {
		override val value: T
			get() = getter.invoke()
	}

	override val entries: Set<Map.Entry<K, T>> = ImmutableSet.copyOf(mValues.map { Entry(it.first, it.second) })
}

class RegistryObjectList<T : IForgeRegistryEntry<T>>(private vararg val getters: RegistryObject<T>) : AbstractList<T>() {
	constructor(getters: List<RegistryObject<T>>) : this(*getters.toTypedArray())

	override val size: Int
		get() = getters.size

	override fun get(index: Int): T {
		return getters[index].get()
	}
}
